local LMP = LibStub("LibMapPins-1.0")

local Harvest = _G["Harvest"]
local pairs = _G["pairs"]

local CallbackManager = Harvest.callbackManager
local Events = Harvest.events


local MapPins = {}
Harvest.mapPins = MapPins

function MapPins:Initialize()
	self.pinTable = {}
	self.hasEntries = false
	self.lastViewedX = -10
	self.lastViewedY = -10
	
	CallbackManager:RegisterForEvent(Events.CHANGED_NODE_HIDDEN_STATE, function(event, map, nodeId, newState)
		self:OnChangedNodeHiddenState(map, nodeId, newState)
	end)
	local callback = function(...) self:OnNodeChangedCallback(...) end
	CallbackManager:RegisterForEvent(Events.NODE_DELETED, callback)
	CallbackManager:RegisterForEvent(Events.NODE_UPDATED, callback)
	CallbackManager:RegisterForEvent(Events.NODE_ADDED, callback)
	
	self.pinManager = LMP.pinManager
	-- register callbacks for events, that affect map pins
	CallbackManager:RegisterForEvent(Events.SETTING_CHANGED, function(...) self:OnSettingsChanged(...) end)

	for index, pinTypeId in pairs( Harvest.PINTYPES ) do
		self:InitializeMapPinType( pinTypeId )
	end

	-- pin type which is always visible to receive the refresh callback
	LMP:AddPinType(
		Harvest.GetPinType( "0" ),
		function() MapPins:PinTypeRefreshCallback() end,
		nil,
		Harvest.GetMapPinLayout( 1 ),
		self.tooltipCreator
	)
end

-- called for each pinTypeId to register the pinType in the lib map pin library.
-- also adds a checkbox for the pinType in the map's filter panel
function MapPins:InitializeMapPinType( pinTypeId )
	local pinType = Harvest.GetPinType( pinTypeId )

	if pinTypeId == Harvest.TOUR then
		LMP:AddPinType(
			pinType,
			HarvestFarm.MapCallback,
			nil,
			Harvest.GetMapPinLayout( pinTypeId ),
			Harvest.tooltipCreator
		)

	else

		LMP:AddPinType(
			pinType,
			function() end,
			nil,
			Harvest.GetMapPinLayout( pinTypeId ),
			Harvest.tooltipCreator
		)
		
		--if not Harvest.GetPinTypeInGroup(pinTypeId) then
		
			local pve, pvp, imperial = LMP:AddPinFilter(
				pinType,
				Harvest.GetLocalization( "pintype" .. pinTypeId ),
				false,
				Harvest.savedVars["settings"].isPinTypeVisible,
				pinTypeId, pinTypeId, pinTypeId
			)
			
			local fun = function(button, visible)
				Harvest.SetMapPinTypeVisible( pinTypeId, visible )
			end
			ZO_CheckButton_SetToggleFunction(pve, fun)
			ZO_CheckButton_SetToggleFunction(pvp, fun)
			ZO_CheckButton_SetToggleFunction(imperial, fun)
		--end
	end
end

function MapPins:OnNodeChangedCallback(event, map, pinTypeId, nodeId)
	local nodeAdded = (event == Events.NODE_ADDED)
	local nodeUpdated = (event == Events.NODE_UPDATED)
	local nodeDeleted = (event == Events.NODE_DELETED)
	
	-- when the heatmap is active, the map pins aren't used
	if Harvest.IsHeatmapActive() then
		return
	end
	
	-- the node isn't on the currently displayed map
	if not self:IsActiveMap(map) then
		return
	end
	
	
	if Harvest.ShouldRefreshPinsOnHarvest() then
		self:RefreshPins(pinTypeId)
	else
		-- refresh a single pin by removing and recreating it
		if not nodeAdded then
			self:RemoveNode(nodeId)
		end
		-- the (re-)creation of the pin is performed
		if not nodeDeleted and not self.mapCache:IsHidden(nodeId) then
			self:AddNode(nodeId)
		end
	end
end

function MapPins:OnChangedNodeHiddenState(map, nodeId, newState)
	if self:IsActiveMap(map) then
		if newState then
			self:RemoveNode(nodeId)
		else
			self:AddNode(nodeId)
		end
	end
end

function MapPins:AddNode(nodeId)
	if not self.mapCache then return end
	if self.pinTable[nodeId] and self.pinTable[nodeId] < 0 then
		self.pinTable[nodeId] = nil
	else
		self.pinTable[nodeId] = self.mapCache.pinTypeId[nodeId]
	end
	self.hasEntries = (next(self.pinTable) ~= nil)
end

function MapPins:RemoveNode(nodeId)
	if not self.mapCache then return end
	self.pinTable[nodeId] = nil
	self.hasEntries = (next(self.pinTable) ~= nil)

	LMP:RemoveCustomPin(Harvest.GetPinType(self.mapCache.pinTypeId[nodeId]), nodeId)
end

-- Clear the queue. Used when the currently viewed map changes or the pins are to be refreshed.
function MapPins:ClearQueue(pinTypeId)
	if not pinTypeId then
		self.pinTable = {}
		self.hasEntries = false
	else
		for nodeId, pinType in pairs(self.pinTable) do
			if pinType == pinTypeId or pinType == -pinTypeId then
				self.pinTable[nodeId] = nil
			end
		end
		self.hasEntries = (next(self.pinTable) ~= nil)
	end
end

function MapPins:ShouldDelay()
	if Harvest.IsDelayedWhenMapClosed() and not ZO_WorldMap_IsWorldMapShowing() then
		return true
	end
	if Harvest.IsDelayedWhenInFight() and IsUnitInCombat("player") then
		return true
	end
	return false
end


function MapPins:RefreshPins(pinTypeId)
	if pinTypeId then
		if pinTypeId == Harvest.TOUR then
			LMP:RefreshPins( Harvest.GetPinType(pinTypeId) )
			return
		end
		
		LMP:RefreshPins(Harvest.GetPinType(pinTypeId))
		self:RemovePins(pinTypeId)

		if not Harvest.IsUpdateQueueEmpty() then
			Harvest.Debug("no pin refresh performed, your data is still being refactored/updated" )
			return
		end

		if Harvest.IsHeatmapActive() then
			return
		end
		
		if Harvest.IsMapPinTypeVisible(pinTypeId) then
			self:AddVisibleNodesOfPinType(pinTypeId)
		end
	else
		for _, pinTypeId in pairs(Harvest.PINTYPES) do
			LMP:RefreshPins(Harvest.GetPinType(pinTypeId))
		end
		LMP:RefreshPins( Harvest.GetPinType( "0" ))
	end
end

function MapPins:UpdateVisibleMapPins()
	if Harvest.IsHeatmapActive() then return end

	local map = Harvest.GetMap()
	local x, y = GetMapPlayerPosition("player")

	self:AddAndRemoveVisblePins(map, x, y)
end

-- If the player moved, some pins enter the visible radius while others leave it.
-- This function updates the queue with new creation/removal commands according to the movement.
function MapPins:AddAndRemoveVisblePins(map, x, y)
	if not self.mapCache then return end
	-- abort if the maps don't match
	-- i.e this queue is used by the compass and the player is viewing a different map than he is currently on.
	if self.mapCache.map ~= map then return end

	local shouldSaveCoords
	shouldSaveCoords = self.mapCache:SetPrevAndCurVisibleNodesToTable(self.lastViewedX, self.lastViewedY, x, y, self.pinTable)

	if shouldSaveCoords then
		self.lastViewedX = x
		self.lastViewedY = y
	end

	self.hasEntries = (next(self.pinTable) ~= nil)
end

function MapPins:RemovePins(pinTypeId)
	-- clear the pin creation queue and remove the pins form the map
	if pinTypeId then
		self:ClearQueue(pinTypeId)
		self.pinManager:RemovePins(Harvest.GetPinType(pinTypeId))
	else
		self:ClearQueue()
		for _, pinTypeId in ipairs(Harvest.PINTYPES) do
			if pinTypeId ~= Harvest.TOUR then
				self.pinManager:RemovePins(Harvest.GetPinType(pinTypeId))
			end
		end
	end
end

function MapPins:SetToMapAndPosition(map, x, y, measurement, zoneIndex)
	if self.mapCache then
		self.mapCache.accessed = self.mapCache.accessed - 1
	end
	self.currentMap = map
	self.mapCache = Harvest.Data:GetMapCache(nil, map, measurement, zoneIndex)
	if not self.mapCache then
		self.hasEntries = false
		return
	end
	self.mapCache.accessed = self.mapCache.accessed + 1
	self.lastViewedX = x
	self.lastViewedY = y
end

---Adds only nodes of the given pinType to the creation queue. Used when only a certain pinType is refreshed.
-- (E.g. the player enabled the pinType on the map)
function MapPins:AddVisibleNodesOfPinType(pinTypeId)
	if not self.mapCache then return end
	Harvest.Data:CheckPinTypeInCache(pinTypeId, self.mapCache)
	
	self.mapCache:GetVisibleNodes(self.lastViewedX, self.lastViewedY, pinTypeId, self.pinTable)
	self.hasEntries = (next(self.pinTable) ~= nil)
end

-- Creates and removes a bunch of queued pins.
function MapPins:PerformActions()
	if not self.hasEntries then
		return
	end
	if self:ShouldDelay() then
		return
	end
	local numNodes = Harvest.GetDisplaySpeed()
	local x, y
	for nodeId, pinTypeId in pairs(self.pinTable) do
		if pinTypeId > 0 then
			x = self.mapCache.localX[nodeId]
			y = self.mapCache.localY[nodeId]
			LMP:CreatePin(Harvest.GetPinType(pinTypeId), nodeId, x, y)
		else
			x = self.mapCache.localX[nodeId]
			y = self.mapCache.localY[nodeId]
			LMP:RemoveCustomPin(Harvest.GetPinType(-pinTypeId), nodeId)
		end
		self.pinTable[nodeId] = nil
		numNodes = numNodes - 1
		if numNodes <= 0 then
			return
		end
	end
	self.hasEntries = false
end

function MapPins:IsActiveMap(map)
	if not self.mapCache then return false end
	return (self.mapCache.map == map)
end

function MapPins:PinTypeRefreshCallback()
	Harvest.Debug("Refresh of pins requested")

	-- stop adding/removing map pins as they belong to the previous map
	self:RemovePins()

	-- data is still being manipulated, better if we don't access it yet
	if not Harvest.IsUpdateQueueEmpty() then
		Harvest.Debug("your data is still being refactored/updated" )
		return
	end

	if Harvest.IsHeatmapActive() then
		Harvest.Debug("pins could not be refreshed" )
		return
	end

	local map, x, y, measurement, zoneIndex = Harvest.GetLocation( true )
	self:SetToMapAndPosition(map, x, y, measurement, zoneIndex)
	for _, pinTypeId in ipairs(Harvest.PINTYPES) do
		if Harvest.IsMapPinTypeVisible(pinTypeId) then
			self:AddVisibleNodesOfPinType(pinTypeId)
		end
	end

end

-- when the debug mode is enabled, this is the description of what happens if the player clicks on a pin.
-- if a pin is clicked, the node is deleted.
local debugHandler = {
	{
		name = MapPins.nameFunction,
		callback = function(pin)
			-- remove this callback if the debug mode is disabled
			if not Harvest.IsDebugEnabled() or IsInGamepadPreferredMode() then
				return
			end
			-- otherwise request the node to be deleted
			local pinType, nodeId = pin:GetPinTypeAndTag()
			local pinTypeId = MapPins.mapCache.pinTypeId[nodeId]
			local nodeIndex = MapPins.mapCache.nodeIndex[nodeId]
			local saveFile = Harvest.Data:GetSaveFile( MapPins.mapCache.map )
			MapPins.mapCache:Delete(nodeId)
			saveFile.data[ MapPins.mapCache.map ][ pinTypeId ][ nodeIndex ] = nil
			
		end,
		show = function() return true end,
		duplicates = function(pin1, pin2) return not Harvest.IsDebugEnabled() end,
		gamepadName = nameFun
	}
}

MapPins.refreshOnSetting = {
	hasVisibleDistance = true,
	visibleDistance = true,
	pinAbovePoi = true,
	heatmapActive = true,
}
function MapPins:OnSettingsChanged(event, setting, ...)
	if self.refreshOnSetting[setting] then
		MapPins:RefreshPins()
	elseif setting == "mapPinTypeVisible" then
		local pinTypeId, visible = ...
		--[[
		local pinTypeGroup = Harvest.GetPinTypeGroup(pinTypeId)
		if pinTypeGroup then
			for _, pinTypeId in pairs(pinTypeGroup) do
				if visible then
					MapPins:RefreshPins(pinTypeId)
				else
					MapPins:RemovePins(pinTypeId)
				end
				LMP:SetEnabled( Harvest.GetPinType(pinTypeId), visible )
			end
		end
		--]]
		if visible then
			MapPins:RefreshPins(pinTypeId)
		else
			MapPins:RemovePins(pinTypeId)
		end
		LMP:SetEnabled( Harvest.GetPinType(pinTypeId), visible )
	elseif setting == "debugModeEnabled" then
		local enabled = ...
		if enabled and not IsInGamepadPreferredMode() then
			Harvest.mapPinClickHandler:Push("debug", debugHandler)
		else
			Harvest.mapPinClickHandler:Pop("debug")
		end
	elseif setting == "pinTypeSize" then
		local pinTypeId, size = ...
		d(...)
		LMP:SetLayoutKey( Harvest.GetPinType( pinTypeId ), "size", size )
		MapPins:RefreshPins( pinTypeId )
	elseif setting == "pinTypeColor" then
		local pinTypeId, r, g, b = ...
		local layout = LMP:GetLayoutKey( Harvest.GetPinType( pinTypeId ), "tint" )
		layout:SetRGB( r, g, b )
		MapPins:RefreshPins( pinTypeId )
	elseif setting == "cacheCleared" then
		local map = ...
		if map == MapPins.currentMap or not map then
			MapPins:RefreshPins()
		end
	end
end

-- description for the map pin tooltips
Harvest.tooltipCreator = {
	creator = function( pin ) -- this function is called when the mouse is over a pin and a tooltip has to be displayed
		if not Harvest.AreDebugMessagesEnabled() then
			local lines = Harvest.GetLocalizedTooltip( pin, MapPins.mapCache )
			for _, line in ipairs(lines) do
				InformationTooltip:AddLine( line )
			end
		else
			local nodeId = pin.m_PinTag
			local pinTypeId = Harvest.GetPinId( LMP.pinManager.customPins[pin.m_PinType].pinTypeString )
			
			if pinTypeId == Harvest.TOUR then
				local lines = Harvest.GetLocalizedTooltip( pin, MapPins.mapCache )
				for _, line in ipairs(lines) do
					InformationTooltip:AddLine( line )
				end
				return
			end
			
			pinTypeId = MapPins.mapCache.pinTypeId[nodeId]
			local nodeIndex = MapPins.mapCache.nodeIndex[nodeId]
			local map = MapPins.mapCache.map
			
			InformationTooltip:AddLine( tostring(nodeIndex) .. ", " .. map)
			
			local saveFile = Harvest.Data:GetSaveFile( map )
			local ace = saveFile.data[ map ][ pinTypeId ][ nodeIndex ]
			
			InformationTooltip:AddLine( tostring(ace) )
		end
	end,
	tooltip = 1
}

-- used by the gamepad UI. The pin name is essentially the gamepad's equivalent of a tooltip
MapPins.nameFunction = function( pin )
	local lines = Harvest.GetLocalizedTooltip( pin, MapPins.mapCache )
	return table.concat(lines, "\n")
end
